/*
 * Server-side request handling
 *
 * Copyright (C) 1998 Alexandre Julliard
 * Copyright (C) 2004 TransGaming Technologies
 */

#include "config.h"
#include "wine/port.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#include <sys/uio.h>
#include <sys/un.h>
#include <unistd.h>

#include "winnt.h"
#include "winbase.h"
#include "wincon.h"
#include "thread.h"
#include "process.h"
#define WANT_REQUEST_HANDLERS
#include "request.h"
#include "server_memory.h"
#include "object.h"

/* Some versions of glibc don't define this */
#ifndef SCM_RIGHTS
#define SCM_RIGHTS 1
#endif

 /* path names for server master Unix socket */
#define CONFDIR    "/.wine"        /* directory for Wine config relative to $HOME */
#define SERVERDIR  "wineserver-"   /* server socket directory (hostname appended) */
#define SOCKETNAME "socket"        /* name of the socket file */

struct master_socket
{
    struct object       obj;         /* object header */
};

static void master_socket_dump( struct object *obj, int verbose );
static void master_socket_poll_event( struct object *obj, int event );
static void master_socket_destroy( struct object *obj );

static const struct object_ops master_socket_ops =
{
    sizeof(struct master_socket),  /* size */
    master_socket_dump,            /* dump */
    no_add_queue,                  /* add_queue */
    NULL,                          /* remove_queue */
    NULL,                          /* signaled */
    NULL,                          /* satisfied */
    NULL,                          /* get_poll_events */
    master_socket_poll_event,      /* poll_event */
    no_get_fd,                     /* get_fd */
#if defined( CONTEXTUAL_FD )
    no_get_ctx_fd,                 /* get_ctx_fd */
#endif
    no_flush,                      /* flush */
    no_get_file_info,              /* get_file_info */
    NULL,                          /* queue_async */
    master_socket_destroy          /* destroy */
};
const struct object_ops* get_this_process_master_socket_ops(void) { return &master_socket_ops; }

/* Per process variables */
struct thread *current = NULL;  /* thread handling the current request */
unsigned int global_error = 0;  /* global error code for when no thread is current */
static struct master_socket *master_socket;  /* the master socket object */

/* socket communication static structures. Only used in wineserver process for master socket  */
static struct iovec myiovec;
static struct msghdr msghdr;

/* End per process variables */

#ifndef HAVE_MSGHDR_ACCRIGHTS
struct cmsg_fd
{
    int len;   /* sizeof structure */
    int level; /* SOL_SOCKET */
    int type;  /* SCM_RIGHTS */
    int fd;    /* fd to pass */
};
static struct cmsg_fd cmsg = { sizeof(cmsg), SOL_SOCKET, SCM_RIGHTS, -1 };
#endif  /* HAVE_MSGHDR_ACCRIGHTS */

/* complain about a protocol error and terminate the client connection */
void fatal_protocol_error( struct thread *thread, const char *err, ... )
{
    va_list args;

    va_start( args, err );
    fprintf( stderr, "%04x: Protocol error: ",  thread ? get_thread_id( thread ) : INVALID_RESOURCE_ID );
    vfprintf( stderr, err, args );
    va_end( args );
    thread->exit_code = 1;
    kill_thread( thread, 1 );
}

/* complain about a protocol error and terminate the client connection */
void fatal_protocol_perror( struct thread *thread, const char *err, ... )
{
    va_list args;

    va_start( args, err );
    fprintf( stderr, "%04x: Protocol error: ",  thread ? get_thread_id( thread ) : INVALID_RESOURCE_ID );
    vfprintf( stderr, err, args );
    perror( " " );
    va_end( args );
    thread->exit_code = 1;
    kill_thread( thread, 1 );
}

/* die on a fatal error */
void fatal_error( const char *err, ... )
{
    va_list args;

    va_start( args, err );
    fprintf( stderr, "wineserver: " );
    vfprintf( stderr, err, args );
    va_end( args );
    exit(1);
}

/* die on a fatal error */
void fatal_perror( const char *err, ... )
{
    va_list args;

    va_start( args, err );
    fprintf( stderr, "wineserver: " );
    vfprintf( stderr, err, args );
    perror( " " );
    va_end( args );
    exit(1);
}

void info_msg( enum wineserver_debug_level importance, const struct thread *thread,
                const char* err, ... )
{
    if( debug_level >= importance )
    {
        va_list args;

        va_start( args, err );
        fprintf( stderr, "%04x:%c ", thread ? get_thread_id( thread ) : 0, is_wineserver ? ':' : ' ' );
        vfprintf( stderr, err, args );
        va_start( args, err );
    }
}

/* allocate the reply data */
void *set_reply_data_size( size_t size )
{
    assert( size <= get_reply_max_size() );
    if (size && !(current->reply_data = mem_alloc_no_init( size ))) size = 0;
    current->reply_size = size;
    return current->reply_data;
}

/* write the remaining part of the reply */
void write_reply( struct thread *thread )
{
    int ret;

    if ((ret = write( thread->reply_fd,
                      (char *)thread->reply_data + thread->reply_size - thread->reply_towrite,
                      thread->reply_towrite )) >= 0)
    {
        if (!(thread->reply_towrite -= ret))
        {
            mem_free( thread->reply_data );
            thread->reply_data = NULL;
            /* sent everything, can go back to waiting for requests */
            change_select_fd( &thread->obj, thread->request_fd, POLLIN );
        }
        return;
    }
    if (errno == EPIPE)
        kill_thread( thread, 0 );  /* normal death */
    else if (errno != EWOULDBLOCK && errno != EAGAIN)
        fatal_protocol_perror( thread, "reply write" );
}

/* send a reply to the current thread */
static void send_reply( union generic_reply *reply )
{
    int ret;

    if (!current->reply_size)
    {
        if ((ret = write( current->reply_fd, reply, sizeof(*reply) )) != sizeof(*reply)) goto error;
    }
    else
    {
        struct iovec vec[2];

        vec[0].iov_base = reply;
        vec[0].iov_len  = sizeof(*reply);
        vec[1].iov_base = current->reply_data;
        vec[1].iov_len  = current->reply_size;

        if ((ret = writev( current->reply_fd, vec, 2 )) < sizeof(*reply)) goto error;

        if ((current->reply_towrite = current->reply_size - (ret - sizeof(*reply))))
        {
            /* couldn't write it all, wait for POLLOUT */
            change_select_fd( &current->obj, current->reply_fd, POLLOUT );
            return;
        }
    }
    if (current->reply_data)
    {
        mem_free( current->reply_data );
        current->reply_data = NULL;
    }
    return;

 error:
    if (ret >= 0)
        fatal_protocol_error( current, "partial write %d\n", ret );
    else if (errno == EPIPE)
        kill_thread( current, 0 );  /* normal death */
    else
        fatal_protocol_perror( current, "reply write" );
}

/* Call a request handler. Returns FALSE if there is no reply, TRUE otherwise */
static int call_req_handler( struct thread *thread, union generic_reply* reply )
{
    enum request req = thread->req->request_header.req;
    int rc = FALSE;

    current = thread;
    current->reply_size = 0;
    clear_error();
    memset( reply, 0, sizeof(reply) );

    if (debug_level) trace_request();

    if (req < REQ_NB_REQUESTS)
    {
        req_handlers[req]( current->req, reply );
        if (current)
        {
            reply->reply_header.error = current->error;
            reply->reply_header.reply_size = current->reply_size;
            if (debug_level) trace_reply( req, reply );
            rc = TRUE;
        }
        return rc;
    }

    fatal_protocol_error( current, "bad request %d\n", req );

    return rc;
}

/* Called from the wineserver only - handles generating the appropriate reply */
static void poll_call_req_handler( struct thread *thread )
{
  union generic_reply reply;

  if( call_req_handler( thread, &reply ) )
     send_reply( &reply );

  current = NULL;
}

/* read a request from a thread */
void read_request( struct thread *thread )
{
    int ret;
    union generic_request req;
#if defined( FAST_VAR_DATA )
    struct __server_iovec iovec;
#endif

    if (!thread->req_toread)  /* no pending request */
    {
        if ((ret = read( get_object_fd_for_server( &thread->obj ), &req,
                         sizeof(req) )) != sizeof(req)) goto error;
        thread->req = &req;
        if (!(thread->req_toread = req.request_header.request_size))
        {
            /* no variable sized data, handle request at once */
            poll_call_req_handler( thread );
            return;
        }
        if (!(thread->req_data = mem_alloc( thread->req_toread )))
            fatal_protocol_error( thread, "no memory for %d bytes request\n", thread->req_toread );
#if defined( FAST_VAR_DATA )
        thread->req_vdata = &iovec;
        thread->req_maxvdata = 1;
        iovec.ptr = thread->req_data;
        iovec.size = thread->req_toread;
#endif
    }

    /* read the variable sized data */
    for (;;)
    {
        ret = read( get_object_fd_for_server( &thread->obj ), ((char *)thread->req_data +
                                     req.request_header.request_size - thread->req_toread),
                    thread->req_toread );
        if (ret <= 0) break;
        if (!(thread->req_toread -= ret))
        {
            poll_call_req_handler( thread );
            mem_free( thread->req_data );
            thread->req_data = NULL;
#if defined( FAST_VAR_DATA )
            thread->req_vdata = NULL;
            thread->req_maxvdata = 0;
#endif
            return;
        }
    }

error:
    if (!ret)  /* closed pipe */
        kill_thread( thread, 0 );
    else if (ret > 0)
        fatal_protocol_error( thread, "partial read %d\n", ret );
    else if (errno != EWOULDBLOCK && errno != EAGAIN)
        fatal_protocol_perror( thread, "read" );
}

/* receive a file descriptor on the process socket */
int receive_fd( struct process *process )
{
    struct send_fd data;
    int fd, ret;

#ifdef HAVE_MSGHDR_ACCRIGHTS
    msghdr.msg_accrightslen = sizeof(int);
    msghdr.msg_accrights = (void *)&fd;
#else  /* HAVE_MSGHDR_ACCRIGHTS */
    msghdr.msg_control    = &cmsg;
    msghdr.msg_controllen = sizeof(cmsg);
    cmsg.fd = -1;
#endif  /* HAVE_MSGHDR_ACCRIGHTS */

    myiovec.iov_base = &data;
    myiovec.iov_len  = sizeof(data);

    ret = recvmsg( get_object_fd_for_server( &process->obj ), &msghdr, 0 );
#ifndef HAVE_MSGHDR_ACCRIGHTS
    fd = cmsg.fd;
#endif

    if (ret == sizeof(data))
    {
        struct thread *thread;

        if (data.tid) thread = get_thread_from_id( data.tid );
        else thread = (struct thread *)grab_object( process->thread_list );

        if (!thread || thread->process != process)
        {
            info_msg( WITH_DEBUG, NULL, "*fd* %d <- %d bad thread id %d\n", data.fd, fd, data.tid );
            close( fd );
        }
        else
        {
            info_msg( WITH_DEBUG, thread, "*fd* %d <- %d\n", data.fd, fd );
            thread_add_inflight_fd( thread, data.fd, fd );
        }
        if (thread) { release_object( thread ); }
        return 0;
    }

    if (ret >= 0)
    {
        if (ret > 0)
            fprintf( stderr, "Protocol error: process %p: partial recvmsg %d for fd\n",
                     process, ret );
        kill_process( process, NULL, 1 );
    }
    else
    {
        if (errno != EWOULDBLOCK && errno != EAGAIN)
        {
            fprintf( stderr, "Protocol error: process %p: ", process );
            perror( "recvmsg" );
            kill_process( process, NULL, 1 );
        }
    }
    return -1;
}

/* send an fd to a client */
int send_client_fd( struct process *process, int fd, handle_t handle )
{
    int ret;

    info_msg( WITH_DEBUG, current, "*fd* %d -> %d\n", handle, fd );

#ifdef HAVE_MSGHDR_ACCRIGHTS
    msghdr.msg_accrightslen = sizeof(fd);
    msghdr.msg_accrights = (void *)&fd;
#else  /* HAVE_MSGHDR_ACCRIGHTS */
    msghdr.msg_control    = &cmsg;
    msghdr.msg_controllen = sizeof(cmsg);
    cmsg.fd = fd;
#endif  /* HAVE_MSGHDR_ACCRIGHTS */

    myiovec.iov_base = (void *)&handle;
    myiovec.iov_len  = sizeof(handle);

    ret = sendmsg( get_object_fd_for_server( &process->obj ), &msghdr, 0 );

    if (ret == sizeof(handle)) return 0;

fprintf( stderr, "Failed to send fd\n" );
    if (ret >= 0)
    {
        if (ret > 0)
            fprintf( stderr, "Protocol error: process %p: partial sendmsg %d\n", process, ret );
        kill_process( process, NULL, 1 );
    }
    else
    {
        fprintf( stderr, "Protocol error: process %p: ", process );
        perror( "sendmsg" );
        kill_process( process, NULL, 1 );
    }
    return -1;
}

/* get current tick count to return to client */
unsigned int get_tick_count(void)
{
    struct timeval t;
    gettimeofday( &t, NULL );
    return (t.tv_sec * 1000) + (t.tv_usec / 1000) - SERVER_GLOBAL_VARIABLE( server_start_ticks );
}

static void master_socket_dump( struct object *obj, int verbose )
{
    struct master_socket *sock = (struct master_socket *)obj;
    assert( obj->ops == master_socket_ops_tag );
    fprintf( stderr, "Master socket fd=%d\n", get_object_fd_for_server( &sock->obj ) );
}

/* handle a socket event */
static void master_socket_poll_event( struct object *obj, int event )
{
    struct master_socket *sock = (struct master_socket *)obj;
    assert( obj->ops == master_socket_ops_tag );

    assert( sock == master_socket );  /* there is only one master socket */

    if (event & (POLLERR | POLLHUP))
    {
        /* this is not supposed to happen */
        fprintf( stderr, "wineserver: Error on master socket\n" );
        release_object( obj );
    }
    else if (event & POLLIN)
    {
        struct sockaddr_un dummy;
        int len = sizeof(dummy);
        int client = accept( get_object_fd_for_server( &master_socket->obj ), (struct sockaddr *) &dummy, &len );
        if (client == -1) return;
        fcntl( client, F_SETFL, O_NONBLOCK );
        create_process( client );
    }
}

/* remove the socket upon exit */
static void socket_cleanup(void)
{
    static int do_it_once;
    if (!do_it_once++) unlink( SOCKETNAME );
}

static void master_socket_destroy( struct object *obj )
{
    assert( obj->ops == master_socket_ops_tag );
    socket_cleanup();
}

/* return the configuration directory ($WINEPREFIX or $HOME/.wine) */
const char *get_config_dir(void)
{
    static char *confdir;
    if (!confdir)
    {
        const char *prefix = getenv( "WINEPREFIX" );
        if (prefix)
        {
            int len = strlen(prefix);

            /* PH: FIXME: Yes use strdup not mem_strdup because it's called so early */
            if (!(confdir = strdup( prefix ))) fatal_error( "out of memory\n" );

            if (len > 1 && confdir[len-1] == '/') confdir[len-1] = 0;
        }
        else
        {
            const char *home = getenv( "HOME" );
            if (!home)
            {
                struct passwd *pwd = getpwuid( getuid() );
                if (!pwd) fatal_error( "could not find your home directory\n" );
                home = pwd->pw_dir;
            }

            if (!(confdir = malloc( strlen(home) + strlen(CONFDIR) + 1 )))
                fatal_error( "out of memory\n" );
            strcpy( confdir, home );
            strcat( confdir, CONFDIR );
        }
    }
    return confdir;
}

/* create the server directory and chdir to it */
static void create_server_dir(void)
{
    char hostname[64];
    char* serverdir;
    const char *confdir = get_config_dir();
    struct stat st;

    if (gethostname( hostname, sizeof(hostname) ) == -1) fatal_perror( "gethostname" );

    if (!(serverdir = mem_alloc( strlen(SERVERDIR) + strlen(hostname) + 1 )))
        fatal_error( "out of memory\n" );

    if (chdir( confdir ) == -1) fatal_perror( "chdir %s", confdir );

    strcpy( serverdir, SERVERDIR );
    strcat( serverdir, hostname );

    if (chdir( serverdir ) == -1)
    {
        if (errno != ENOENT) fatal_perror( "chdir %s", serverdir );
        if (mkdir( serverdir, 0700 ) == -1) fatal_perror( "mkdir %s", serverdir );
        if (chdir( serverdir ) == -1) fatal_perror( "chdir %s", serverdir );
    }
    if (stat( ".", &st ) == -1) fatal_perror( "stat %s", serverdir );
    if (!S_ISDIR(st.st_mode)) fatal_error( "%s is not a directory\n", serverdir );
    if (st.st_uid != getuid()) fatal_error( "%s is not owned by you\n", serverdir );
    if (st.st_mode & 077) fatal_error( "%s must not be accessible by other users\n", serverdir );

    SERVER_GLOBAL_VARIABLE( serverdir ) = serverdir;
}

/* open the master server socket and start waiting for new clients */
void open_master_socket(void)
{
    struct sockaddr_un addr;
    int fd, slen;

    /* make sure no request is larger than the maximum size */
    assert( sizeof(union generic_request) == sizeof(struct request_max_size) );
    assert( sizeof(union generic_reply) == sizeof(struct request_max_size) );

    create_server_dir();
    if ((fd = socket( AF_UNIX, SOCK_STREAM, 0 )) == -1) fatal_perror( "socket" );
    addr.sun_family = AF_UNIX;
    strcpy( addr.sun_path, SOCKETNAME );
    slen = sizeof(addr) - sizeof(addr.sun_path) + strlen(addr.sun_path) + 1;
#ifdef HAVE_SOCKADDR_SUN_LEN
    addr.sun_len = slen;
#endif
    if (bind( fd, (struct sockaddr *)&addr, slen ) == -1)
    {
        if ((errno == EEXIST) || (errno == EADDRINUSE))
            exit(0);  /* pretend we succeeded to start */
        else
            fatal_perror( "bind" );
    }
    if( atexit( socket_cleanup ) ) fatal_perror( "master socket" );

    chmod( SOCKETNAME, 0600 );  /* make sure no other user can connect */
    if (listen( fd, 5 ) == -1) fatal_perror( "listen" );

    if (!(master_socket = alloc_object( master_socket_ops_tag, fd )))
        fatal_error( "out of memory\n" );
    set_select_events( &master_socket->obj, POLLIN, NULL );

    /* setup msghdr structure constant fields */
    msghdr.msg_name    = NULL;
    msghdr.msg_namelen = 0;
    msghdr.msg_iov     = &myiovec;
    msghdr.msg_iovlen  = 1;

    /* init startup ticks */
    SERVER_GLOBAL_VARIABLE( server_start_ticks ) = get_tick_count();

#if !defined( NO_SHARED_MEMORY )
    open_fd_server_socket();
#endif

    /* go in the background if requested */
    if( !wineserver_is_daemon ) return;

    switch(fork())
    {
    case -1:
        fatal_perror( "fork" );
    case 0:
        setsid();
        break;
    default:
        _exit(0);  /* do not call atexit functions */
    }
}

/* close the master socket and stop waiting for new clients */
void close_master_socket(void)
{
    /* if a new client is waiting, we keep on running */
    if (!check_select_events(get_object_fd_for_server( &master_socket->obj ), POLLIN ))
    {
        release_object( master_socket );
        close_fd_server_socket();
        destroy_distributed_command();

    }
}

/* lock/unlock the master socket to stop accepting new clients */
void lock_master_socket( int locked )
{
    set_select_events( &master_socket->obj, locked ? 0 : POLLIN, NULL );
}



#if !defined( FAST_VAR_DATA )
/* PH: FIXME: Hmmm. */
struct __server_iovec
{
    const void  *ptr;
    unsigned int size;
};
#endif

#define __SERVER_MAX_DATA 4

struct __server_request_info
{
    union
    {
        union generic_request req;    /* request structure */
        union generic_reply   reply;  /* reply structure */
    } u;
    size_t                size;       /* size of request structure */
    unsigned int          data_count; /* count of request data pointers */
    void                 *reply_data; /* reply data pointer */
    struct __server_iovec data[__SERVER_MAX_DATA];  /* request variable size data */
};


/* Entry point for shared memory server */
int shared_server_call( thread_id_t thread_id,
                        const struct __server_request_info *req,
                        union generic_reply* reply,
                        void* variable_reply_data )
{
  int reply_generated = 0;
  struct thread* thread;

  acquire_shm_server_semaphore_exclusive();

  thread = get_thread_from_id(thread_id);

  /* Copy the request */
  thread->req = &req->u.req;

#if defined( FAST_VAR_DATA )
  thread->req_vdata = req->data;
  thread->req_maxvdata = req->data_count;
#else
  /* Simulate "reading the variable data" */
  if( req->u.req.request_header.request_size )
  {
    /* PH: FIXME: If this stuff wasn't passed knowing it was iovec'd it'd save this alloc&memcpy */
    char* data_ptr = thread->req_data = mem_alloc_no_init( req->u.req.request_header.request_size );
    int i;

    for( i = 0; i < req->data_count; i++ )
    {
      memcpy( data_ptr, req->data[i].ptr, req->data[i].size );
      data_ptr += req->data[i].size;
    }
  }
#endif

  /* Make the call */
  reply_generated = call_req_handler( thread, reply );

  thread->req = NULL;
  if( thread->req_data ) mem_free( thread->req_data );
  thread->req_data = NULL;
                                                                                                                                                                        
  /* Copy the variable portion of the reply */
  if( current->reply_size )
  {
    memcpy( variable_reply_data, current->reply_data, current->reply_size );
    mem_free( current->reply_data );
    current->reply_data = NULL;
  }

  current = NULL;

  release_object( thread );  

  release_shm_server_semaphore();

  return reply_generated;
}

int shared_server_complete_poll( thread_id_t thread_id, const void* cookie,
                                 struct shm_pollfd* shm_pollfd, unsigned int size,
                                 unsigned int hits, int signaled )
{
  struct thread* thread;
  int ret = 0;

  acquire_shm_server_semaphore_exclusive();

  // PH: FIXME: What if id doesn't convert?
  thread = get_thread_from_id( thread_id );

  current = thread;

  info_msg( WITH_DEBUG, thread,
              "shared_server_complete_poll called for cookie=%p. hits=%d state=0x%x\n",
              cookie, hits, signaled );

  if( signaled == STATUS_TIMEOUT )
  {
     int timeout_calls;

     if( is_cookie_satisfied( thread, cookie ) )
     {
         /* We're satisfied before we even processed the timeout. This means that
          * there was a race. With this race another thread satisfied the wait condition
          * after poll timed out in the client thread but before we entered the shm
          * wineserver for this call.
          */
         ret = 1;
     }
     else
     {
         timeout_calls = shm_handle_timeout();
 
         /* There should only be 1 timeout outstanding on the thread */
         if( timeout_calls != 1 ) fprintf( stderr, "Hmm. %d timeouts this round\n", timeout_calls );

         /* We have woken ourselves up. Check for the proper timeout response */
         ret = 1;
     }
  }
  else
  {
     shm_poll_hit( shm_pollfd, hits, size );

     /* Wait for a wait_fd reponse in client? */
     ret = is_cookie_satisfied( thread, cookie );
  }

  current = NULL;

  info_msg( WITH_DEBUG, thread, "shared_server_complete_poll cookie=%p returns with %d\n",
                      cookie, ret );

  release_object( thread );

  release_shm_server_semaphore();

  return ret;
}

